require 'timer'

AI_STAND = 1
AI_WALK = 2
AI_FOLLOW = 3
AI_ENTER = 4
AI_SPAWN = 5
AI_GO_DIR = 6
AI_START_GO_CORNER = 7
AI_GO_CORNER = 8
AI_LEAVE = 9

ai = {}

local ai_prob = {
	[TROLL] = {AI_FOLLOW},
	[TANK] = {AI_FOLLOW, AI_WALK, AI_WALK, AI_STAND, AI_STAND},
	[PLANE] = {AI_GO_DIR},
	[BOMB_PLANE] = {AI_GO_DIR},
	[MINE] = {AI_STAND},
	[MONK] = {AI_GO_DIR},
	[BOSS1] = {AI_STAND, AI_STAND, AI_STAND, AI_START_GO_CORNER}
}

local function move_to(from, to, vel)
	local vec = to - from
	vec:normalize_inplace()
	return vec * vel
end


local function box(x, y, w, h)
	return {pos = vector(x, y), size = vector(w, h)}
end

local corner_w, corner_h = 200, 200

local function box_corner(x, y)
	return box(x, y, corner_w, corner_h)
end

local corners = {
	box_corner(0, 0),
	box_corner(600, 0),
	box_corner(600, 400),
	box_corner(0, 400)
}

function ai.create(type, c)
	local obj = {
		type = type,
		state = AI_SPAWN,
		lock_state = true,
		vel = vector(),
		timer = timer.create(),
		obj = c
	}
	setmetatable(obj, {__index = ai})
	if obj.type == MINE then
		obj.lock_state = false
		obj:set_state()
	end

	return obj
end

function ai:set_state()
	if self.lock_state then return end
	
	if self.type == BOSS1 and not self.cur_corner then
		for i = 1, 4 do
			if objs_collide(self.obj, corners[i]) then
				self.cur_corner = i
				break
			end
		end
	end

	if self.state ~= AI_GO_CORNER then
		self.state = ai_prob[self.type][math.random(1, #ai_prob[self.type])]
		if self.state == AI_WALK then
			self.obj.vel = self.obj.speed * self:aim(self:vec_to(vector(GAME_W / 2, GAME_H / 2)), 1.5 * math.pi)
		end
	end
	
	self.timer:add_task(function() self:set_state() end, math.random() / 2 + 0.1)
end

-- leave the screen using the shortest way
function ai:leave()
	local dir = 0
	local distance = GAME_W - self.obj.pos.x
	local min_distance = distance
	
	distance = self.obj.pos.y + self.obj.size.y
	if distance < min_distance then
		min_distance = distance
		dir = 1
	end
	
	distance = self.obj.pos.x + self.obj.size.x
	if distance < min_distance then
		min_distance = distance
		dir = 2
	end
	
	distance = GAME_H - self.obj.pos.y
	if distance < min_distance then
		min_distance = distance
		dir = 3
	end

	local goal = self.obj.pos:clone()
	
	if dir == 0 then
		goal.x = GAME_W
	elseif dir == 1 then
		goal.y = - self.obj.size.y
	elseif dir == 2 then
		goal.x = - self.obj.size.x
	else
		goal.y = GAME_H
	end
	
	goal = goal - self.obj.pos
	goal:normalize_inplace()
	self.obj.vel = self.obj.speed * goal
	self.obj.stays_inside = false

	self.state = AI_LEAVE
	self.lock_state = true
end

function ai:spawn()
	self.state = AI_SPAWN
	self.lock_state = true
end

function ai:aim(vec, variation)
	local dir = vec:normalized()
	local angle = math.atan2(dir.y, dir.x)
	angle = angle + (math.random() - 0.5) * variation
	return vec:len() * vector(math.cos(angle), math.sin(angle))
end

function ai:vec_to(vec)
	return (vec - self.obj.pos - self.obj.size / 2):normalize_inplace()
end

function ai:vec_to_player()
	return self:vec_to(game.player.pos + game.player.size / 2)
end

function ai:follow()
	if self.state == AI_FOLLOW then
		self.obj.vel = self.obj.speed * self:aim(self:vec_to_player(), 0.6)
	end
	self.timer:add_task(function() self:follow() end, math.random() / 2 + 0.2)
end

function ai:update(dt)
	self.timer:update(dt)
	if self.state == AI_STAND then
		if self.obj.vel:len() > self.obj.speed * dt * 4 then
			self.obj.vel = self.obj.vel - self.obj.vel:normalized() * self.obj.speed * 4 * dt
		end
	elseif self.state == AI_SPAWN then
		local goal = self.obj.pos:clone()
		if self.obj.pos.x < 0 then
			goal.x = 0
		elseif self.obj.pos.x + self.obj.size.x > GAME_W then
			goal.x = GAME_W - self.obj.size.x
		end
		if self.obj.pos.y < 0 then
			goal.y = 0
		elseif self.obj.pos.y + self.obj.size.y > GAME_H then
			goal.y = GAME_H - self.obj.size.y
		end
		goal = goal - self.obj.pos
		goal:normalize_inplace()
		self.obj.vel = self.obj.speed * goal

		self.state = AI_ENTER

		if self.type == PLANE  then
			if self.obj.vel.x > 0 then
				self.obj.img = get_image_raw("plane.png")
			else
				self.obj.img = get_image_raw("plane_m.png")
			end
		end
		if self.type == BOMB_PLANE  then
			if self.obj.vel.x > 0 then
				self.obj.img = get_image_raw("copter.png")
			else
				self.obj.img = get_image_raw("copter_m.png")
			end
		end
	elseif self.state == AI_ENTER then
		if self.obj:inside_screen() then
			if self.type ~= PLANE and self.type ~= BOMB_PLANE and
				self.type ~= MONK then
				self.obj.stays_inside = true
			end
			
			if self.type == MONK then
				self.timer:add_task(function() self.obj.vel = -self.obj.vel end, 0.8)
			end

			self.lock_state = false
			self:set_state()
			self:follow()
		end
	elseif self.state == AI_GO_DIR then
		if self.obj:outside_screen() then
			game:delete_clown(self.obj.id, true)
		end
	elseif self.state == AI_START_GO_CORNER then
		local corner = self.cur_corner
		while corner == self.cur_corner do
			corner = math.random(4)
		end
		self.state = AI_GO_CORNER		

		self.last_corner = self.cur_corner
		self.cur_corner = corner
		self.obj.vel = move_to(self.obj.pos, corners[corner].pos, self.obj.speed)
		self.shoot_ammo = math.random(15, 20)
		self.next_shoot = math.random() / 50 + self.timer:get_time()
		self.obj.last_shot = self.timer:get_time() + 3.5
	elseif self.state == AI_GO_CORNER then
		if self.shoot_ammo > 0 then
			for i = 1, 4 do
				if i ~= self.last_corner and i ~= self.cur_corner and self.timer:get_time() > self.next_shoot and
						self.last_shot ~= i then
					play_sound("laser2.wav")
					game:add_shot(
						self.obj.pos + self.obj.size/2,
						self:aim(self:vec_to(corners[i].pos + corners[i].size/2), 0.6),
						self.obj.shoot)
					self.shoot_ammo = self.shoot_ammo - 1
					self.next_shoot = self.timer:get_time() + 0.10 + math.random() / 20
					self.last_shot = i
				end
			end
		end

		if obj_inside(self.obj, corners[self.cur_corner]) then
			self.obj.vel = vector(0,0)
			self.state = AI_STAND
		end
	elseif self.state == AI_LEAVE then
		if game:out_of_area(self.obj) then
			self.obj.vel = vector(0,0)
		end
	end

	if self.obj:inside_screen() and
		self.obj.shoot and self.obj.last_shot < self.timer:get_time() then

		if self.obj.shoot == DROP_MINE then
			play_sound("miina.wav")
			game:add_clown(
				self.obj.pos + self.obj.size/2,
				MINE)
		else
			play_sound("laser2.wav")
			game:add_shot(
				self.obj.pos + self.obj.size/2,
				self:aim(self:vec_to_player(), 0.6),
				self.obj.shoot)
		end
		local t = creature_types[self.type]
		self.obj.last_shot = self.timer:get_time() + math.random() * (t.shoot_max_delay - t.shoot_min_delay)
			+ t.shoot_min_delay
	end

end

