require 'hump/vector'
require 'shot'
require 'images'
require 'animation'

TROLL = 1
TANK = 2
PLANE = 3
BOMB_PLANE = 4
MINE = 5
MONK = 6
BOSS1 = 7

require 'ai'

creature = {}

creature_types = {
	[TROLL] = {
		size = vector(26, 40),  -- x2
		img = "ropo.png",
		img_pos = vector(-12, -14),
		img_scale = 2,
		speed = 100,
		hp = 4,
		level = 10,
		strength = 15,
		points = 10,
		mass = 0.5,
	},
	[TANK] = {
		size = vector(57, 24),  -- 3x
		img = "tanc.png",
		img_pos = vector(-6, -15),
		img_scale = 3,
		speed = 60,
		hp = 10,
		shoot = SLOW_PLASMA_BALL,
		mass = 1,
		level = 15,
		strength = 70,
		points = 50,
		shoot_min_delay = 1,
		shoot_max_delay = 3.5,
	},
	[PLANE] = {
		size = vector(62, 14), -- x2
		speed = 350,
		hp = 8,
		mass = 10,
		level = 20,
		strength = 30,
		points = 50,
		img_pos = vector(-2, -10),
		img_scale = 2,
	},
	[BOMB_PLANE] = {
		size = vector(40, 38), -- x2
		speed = 200,
		hp = 10,
		mass = 10,
		level = 20,
		strength = 70,
		points = 80,
		img_pos = vector(-28, -14),
		img_scale = 2,
		shoot = DROP_MINE,
		shoot_min_delay = 0.7,
		shoot_max_delay = 1.4,
	},
	[MINE] = {
		size = vector(44, 24), -- x2
		img = "mine.png",
		img_pos = vector(-6, -10),
		img_scale = 2,
		speed = 1,
		hp = 2,
		mass = 10,
		level = 20,
		strength = 5,
		points = 5,
	},
	[MONK] = {
		size = vector(30, 39), -- x3
		speed = 150,
		hp = 8,
		mass = 10,
		level = 20,
		strength = 40,
		points = 40,
		img_pos = vector(-15, -9),
		img_scale = 3,
		anim_data = animation_data.new(
			get_image("monk.png"),
			20, 16, {
				{name = "right", frame_count = 1},
				{name = "left", frame_count = 1}
			})
	},
	[BOSS1] = {
		size = vector(192, 192), -- x6
		img = "boss1.png",
		img_pos = vector(0, 0),
		img_scale = 6,
		speed = 100,
		hp = 400,
		level = 10,
		strength = 10e6,
		points = 1000,
		mass = 10,
		shoot = SLOW_PLASMA_BALL,
		shoot_min_delay = 0.3,
		shoot_max_delay = 0.7,
	},
}

creature_levels = {}
for _, idx in ipairs( { TROLL, TANK, PLANE, BOMB_PLANE, MONK } ) do
	local ctype = creature_types[idx]
	if not creature_levels[ctype.level] then
		creature_levels[ctype.level] = {}
	end
	
	table.insert(creature_levels[ctype.level], idx)
end

explosion_anim_data = animation_data.new(
	get_image("pum.png"),
	64, 64, {
		{name = "explode", frame_count = 5, fps = 10 }
	})

local ind_l = get_image_raw("indicator_l.png")
local ind_r = get_image_raw("indicator_r.png")
local ind_u = get_image_raw("indicator_u.png")
local ind_d = get_image_raw("indicator_d.png")

function is_boss(type)
	if type == BOSS1 then
		return true
	end
	return false
end

function creature.new(pos, type)
	local delay = 0
	if creature_types[type].shoot_min_delay then
		delay = 1 + 
			creature_types[type].shoot_min_delay + math.random() 
			* (creature_types[type].shoot_max_delay - creature_types[type].shoot_min_delay)
	end

	local c = {
		pos = pos,
		initial_pos = pos:clone(),
		vel = vector(),
		type = type,
		knockback = vector(),
		last_shot = delay,
		hp = creature_types[type].hp,
		size = creature_types[type].size,
		speed = creature_types[type].speed,
		shoot = creature_types[type].shoot,
		mass = creature_types[type].mass,
		img = get_image_raw(creature_types[type].img),
		img_pos = creature_types[type].img_pos,
		img_scale = creature_types[type].img_scale,
	}

	if creature_types[type].anim_data then
		c.anim = animation.new(creature_types[type].anim_data)
		c.anim:set_mode("left")
	end

	c.ai = ai.create(type, c)
	setmetatable(c, {__index = creature})
	return c
end


function creature:hit(shot)
	local dmg = shot.dmg

	self.hp = self.hp - dmg
	if self.hp <= 0 then 
		self:die()
	else
		play_sound("tsut.wav")
	end
	if self:inside_screen() then
		self.knockback = self.knockback + shot.vel:normalized() * 10 / self.mass
	end
end

function creature:die()
	if not self.death_anim then
		self.hp = 0
		game.timer:add_task(function() game:delete_clown(self.id) end, 0.5)
		self.death_anim = animation.new(explosion_anim_data)
		self.death_anim:set_mode("explode", false)
		play_sound("putum.wav")
	end
end

function creature:dead()
	return self.hp <= 0
end

function creature:leave()
	self.ai:leave()
end

function creature:move_knockback(dt)
	if self.knockback:len() > self.speed * dt then
		self.pos = self.pos + self.knockback:normalized() * self.speed * dt / self.mass
		self.knockback = self.knockback - self.knockback:normalized() * self.speed * dt
	end
end

function creature:update(dt)
	if not self:dead() then 
		self.ai:update(dt)
		self:move_knockback(dt)

		self.pos = self.pos + self.vel * dt
	end
	if self.anim then
		self.anim:animate(dt)
	end
	if self.death_anim then
		self.death_anim:animate(dt)
	end

	if self.type == MONK then
		if self.vel.x < 0 then
			self.anim:set_mode("left")
		else
			self.anim:set_mode("right")
		end
	end
end

function creature:draw()
	if self:outside_screen() and self.ai.state ~= AI_GO_DIR then
		local img = ind_d
		local img_pos = self.pos:clone()
		
		if img_pos.x < 5 then
			img_pos.x = 5
			img = ind_l
		elseif img_pos.x + ind_d:getHeight()*SCALE > GAME_W then
			img_pos.x = GAME_W - ind_d:getHeight()*SCALE - 5
			img = ind_r
		end
		if img_pos.y < 5 then
			img_pos.y = 5
			img = ind_u
		elseif img_pos.y + ind_d:getHeight()*SCALE > GAME_H then
			img_pos.y = GAME_H - ind_d:getHeight()*SCALE - 5
			img = ind_d
		end
		
		love.graphics.draw(img, img_pos.x, img_pos.y, 0, SCALE, SCALE)
	else
		if self:dead() then
			self.death_anim:draw(math.ceil(self.pos.x) + self.size.x/2 - 32, 
				math.ceil(self.pos.y) + self.size.y/2 - 32)
		else
			if self.img then
				local img = self.img
				love.graphics.draw(
					img,
					math.ceil(self.pos.x + self.img_pos.x),
					math.ceil(self.pos.y + self.img_pos.y),
					0, self.img_scale, self.img_scale, 0, 0)
			elseif self.anim then
				-- img_pos is not added to animations
				self.anim:draw(math.ceil(self.pos.x) + self.size.x/2 - 32, 
					math.ceil(self.pos.y) + self.size.y/2 - 32,
					self.img_scale, self.img_scale)
			end
		end
	end
end

function creature:inside_screen()
	return (self.pos.x >= 0 and self.pos.y >= 0 and
		self.pos.x + self.size.x <= GAME_W and
		self.pos.y + self.size.y <= GAME_H)
end

function creature:outside_screen()
	return (self.pos.x > GAME_W or self.pos.y > GAME_H or
		self.pos.x + self.size.x < 0 or
		self.pos.y + self.size.y < 0)
end
