---
-- Copy an object.
--
-- @param obj The table to copy.
-- @param deep Whether all subtables and subtables' subtables etc. should be 
-- 			copied.
-- @return The new table.
function copy(obj, deep, saved, ret)
	ret = ret or {}
	saved = saved or {}
	for k, v in pairs(obj) do
		if deep and type(v) == "table" then
			if not saved[v] then
				saved[v] = {}
				ret[k] = copy(v, deep, saved, saved[v])
			else
				ret[k] = saved[v]
			end
		else
			ret[k] = v
		end
	end
	setmetatable(ret, getmetatable(obj))
	return ret
end

function sign(x)
	if x > 0 then
		return 1
	elseif x == 0 then
		return 0
	else 
		return -1
	end
end

function replicate(n, x)
	local t = {}
	for i = 1, n do
		table.insert(t, copy(x, true))
	end
	return t
end

function linear(a, b, c)
	local t = {}
	local c = c or 1
	if a > b then c = -c end 
	for i = a, b, c do
		table.insert(t, i)
	end
	return t
end

function reverse(t)
	local t2 = {}
	local size = table.maxn(t)
	for i = 1, math.ceil(size / 2) do 
		t2[i] = t[size - i + 1]
		t2[size -i + 1] = t[i]
	end
	return t2
end

function concat(...)
	local t = {}
	for k, ar in ipairs(arg) do
		for i, v in ipairs(ar) do
			table.insert(t, v)
		end
	end
	return t
end

function merge(to, from, overwrite, lookup)
	local overwrite = overwrite or false

	if lookup and type(from) == "table" then
		if lookup[from] then return end
		lookup[from] = true
	end

	for k, v in pairs(from) do
		if not overwrite then
			if type(v) == "table" and type(to[k]) == "table" then
				merge(to[k], v, overwrite, lookup)
			else
				if not to[k] then
					to[k] = v
				end
			end
		else
			if type(v) == "table" and type(to[k]) == "table" then
				merge(to[k], v, overwrite, lookup)
			else
				to[k] = v
			end
		end
	end
end

function merge_complex(to, from, overwrite)
	merge(to, from, overwrite, {})
end

function merge_objs(to, from, lookup)
	local lookup = lookup or {}
	local overwrite = overwrite or false

	if lookup and type(from) == "table" then
		if lookup[from] then return end
		lookup[from] = true
	end

	for k, v in pairs(from) do
		if type(v) == "table" and type(to[k]) == "table" then
			if not v.class and not to[k].class then 
				merge(to[k], v, lookup)
			end
		else
			to[k] = v
		end
	end
end

function map(t, f)
	local a = {}
	for i, v in ipairs(t) do
		a[i] = f(v)
	end
	return a
end

function dofile(file)
	if love.filesystem.exists(file) then
		local chunk = love.filesystem.load(file)
		return chunk()
	else
		print("dofile: file " .. file .. " doesn't exist")
	end
end

function find(t, s)
	for k, v in pairs(t) do
		if v == s then
			return k
		end
	end
end

function once(f)
	local called = false
	return function()
		if not called then
			f()
			called = true
		end
	end
end

function write_string(f, string)
	local string = string:gsub("\"", "\\\"")
	string = string:gsub("\n", "\\n")
	string = string:gsub("\r", "\\r")

	f:write("\"" .. string .. "\"")
end

function write_key(f, key)
	-- key
	if type(key) == "string" then
		f:write("[")
		write_string(f, key)
		f:write("]")
	elseif type(key) == "number" then
		f:write("[" .. key .. "]")
	end
end


--[[
-- file format:
--
-- local tables = {}
-- tables[n] = {stuff}
--
-- tables[1] = {a = {[2]=1, b={1}}
-- return tables       ^    ^
--                     |    |
-- reference to tables[1]   |
--                 value 1 -/ 
--]]

function write_table(f, val, tables, lookup)
	local i = #tables
	tables[i+1] = val
	lookup[val] = i+1

	-- first serialize subtables
	for k, v in pairs(val) do
		if type(v) == "table" and not lookup[v] then
			write_table(f, v, tables, lookup)
		end
	end

	-- value
	f:write("tables[" .. (i+1) .. "]=\n")
	f:write("{\n")
	for k, v in pairs(val) do
		if type(v) == "table" or type(v) == "string" or type(v) == "number" or
				type(v) == "boolean" then

			write_key(f, k)
			f:write("=")

			f:write("{")
			if type(v) == "table" then
				f:write("[2]=" .. lookup[v])
			elseif type(v) == "string" then
				write_string(f, v)
			elseif type(v) == "boolean" or type(v) == "number" then
				f:write(tostring(v))
			end
			f:write("}, \n")
		end
	end
	f:write("}\n")
end

function save_table(f, val)
	f:write("local tables = {}\n")
	local tables = {}
	local lookup = {}
	write_table(f, val, tables, lookup)
	f:write("return tables")
end

function load_table(table)
	local f = table
	-- deference subtables
	for i = 1, #f do
		local table = f[i]
		for k, v in pairs(table) do
			if v[2] then
				table[k] = f[v[2]]
			else
				table[k] = v[1]
			end

		end
	end
	return f[1]
end
